import BayesBoom.boom as boom
import numpy as np
import pandas as pd
from pandas.api.types import (
    is_numeric_dtype, is_categorical_dtype, is_object_dtype, is_bool_dtype
)
import pickle
import matplotlib.pyplot as plt
import seaborn as sns
import time

from .data_table import to_data_table, to_data_frame


class AutoClean:
    """
    A model for filling in missing values and correcting errors in a data
    table.

    NOTE: This is experimental code.  Its reliability is by no means guaranteed.
    """

    def __init__(self):
        self._dtypes = None
        self.coefficients = None
        self.residual_variance = None
        self.atom_probs = None
        self.atom_error_probs = None
        self.level_probs = None
        self.level_observation_probs = None
        self.mixing_distribution = None
        self._numeric_colnames = None
        self._categorical_colnames = None
        self._atoms = None
        self._atom_prior = {}
        self._atom_error_prior = {}
        self._level_prior = {}
        self._level_observation_prior = {}

    def __getstate__(self):
        state = {
            "dtypes": self._dtypes,
            "coefficients": self.coefficients,
            "residual_variance": self.residual_variance,
            "atom_probs": self.atom_probs,
            "atom_error_probs": self.atom_error_probs,
            "level_probs": self.level_probs,
            "level_observation_probs": self.level_observation_probs,
            "mixing_distribution": self.mixing_distribution,
            "numeric_colnames": self._numeric_colnames,
            "categorical_colnames": self._categorical_colnames,
            "atoms": self._atoms,
            "atom_prior": self._atom_prior,
            "atom_error_prior": self._atom_error_prior,
            "levels": self._levels,
            "level_prior": self._level_prior,
            "level_observation_prior": self._level_observation_prior,
        }
        return state

    def __setstate__(self, state):
        self._dtypes = state["dtypes"]
        self.coefficients = state["coefficients"]
        self.residual_variance = state["residual_variance"]
        self.atom_probs = state["atom_probs"]
        self.atom_error_probs = state["atom_error_probs"]
        self.level_probs = state["level_probs"]
        self.level_observation_probs = state["level_observation_probs"]
        self.mixing_distribution = state["mixing_distribution"]
        self._numeric_colnames = state["numeric_colnames"]
        self._categorical_colnames = state["categorical_colnames"]
        self._atoms = state["atoms"]
        self._atom_prior = state["atom_prior"]
        self._atom_error_prior = state["atom_error_prior"]
        self._level_prior = state["level_prior"]
        self._level_observation_prior = state["level_observation_prior"]
        self._levels = state["levels"]

    def save(self, filename):
        """
        Save the object to a pickle file.
        """
        with open(filename, "wb") as f:
            pickle.dump(self, f)

    @classmethod
    def load(cls, filename):
        """
        Load a previously fit model from a pickle file.
        """
        with open(filename, "rb") as f:
            return pickle.load(f)   # nosec

    @property
    def nclusters(self):
        if self._model is None:
            return 0
        return self._model.nclusters

    def train_model(self, data: pd.DataFrame, nclusters: int, niter: int,
                    ping: int = 0, checkpoint_filename: str = ""):
        self._start_time = time.time()
        if self._dtypes is None:
            self._dtypes = data.dtypes

        if self._atoms is None:
            self._numeric_colnames = []
            self._categorical_colnames = []
            self._atoms, self._levels = self.find_atoms(data)

        atoms_arg = [boom.Vector(np.array(self._atoms[vname]))
                     for vname in self._numeric_colnames]
        table_arg = to_data_table(data)
        print("creating model object")
        self._model = boom.MixedDataImputerWithErrorCorrection(
            nclusters, table_arg, atoms_arg, boom.GlobalRng.rng)

        print("setting prior")
        self._set_prior()

        print("allocating space")
        self._allocate_space(niter)

        print("about to start mcmc")
        for i in range(niter):
            if (ping > 0) and (i % ping == 0):
                sep = "=-=-=-=-=-=-=-=-="
                print(f"{sep} {time.asctime()} Iteration {i} {sep}")
            self._model.sample_posterior()
            self._record_draws(i)
            if i % niter == 0 and checkpoint_filename != "":
                self.save(checkpoint_filename)

        self._end_time = time.time()

    def find_atoms(self, data: pd.DataFrame):
        """
        Find the numeric atoms and categorical levels to be modeled.
        """
        self._dtypes = data.dtypes
        atoms_dict = {}
        levels_dict = {}
        for i in range(data.shape[1]):
            vname = data.columns[i]
            dt = self._dtypes[i]

            if is_numeric_dtype(dt):
                variable = data.iloc[:, i]
                counts = variable.value_counts().sort_values(ascending=False)
                number_observed = counts.sum()
                atom_indicator = counts > 0.05 * number_observed
                atoms = counts[atom_indicator].index.tolist()
                if len(atoms) > 3:
                    atoms = atoms[:3]
                atoms_dict[vname] = atoms
                self._numeric_colnames.append(data.columns[i])

            elif (
                    is_categorical_dtype(dt)
                    or is_object_dtype(dt)
                    or is_bool_dtype(dt)
            ):
                # TODO: put in some cardinality protections.
                levels = data.iloc[:, i].value_counts()
                levels_dict[vname] = levels
                self._categorical_colnames.append(data.columns[i])

            else:
                raise Exception(
                    "Only categorical or numeric types are supported.")

        return atoms_dict, levels_dict

    def set_atom_prior(self, variable_name: str, prior_counts: np.ndarray):
        """
        Set a constrained (0's are allowed) Dirichlet prior distribution for the
        frequency of "true values" for a numeric variable.

        Args:
          variable_name: The name of the numeric variable that the prior
            describes.
          prior_counts: A 1-d array containing the prior counts.  If the
            variable has k atoms then the array needs k+1 elements, with the
            last one corresponding to the continuous portion of the model.
            Non-positive entries in 'prior_counts' signal that the
            corresponding element has zero probability of being the true value.
        """
        self._atom_prior[variable_name] = prior_counts.ravel()
        if np.sum(prior_counts > 0) == 0:
            raise Exception("At least one prior count must be positive.")

    def set_atom_error_prior(self,
                             variable_name: str,
                             prior_counts: np.ndarray):
        """
        Set a prior distribution on the conditional probability of observing an
        atom in a numeric field, given the true value.

        Args:
          variable_name: The name of the numeric varible described by the
            prior.
          prior_counts: A 2-d numpy array.  If the variable has k atoms then a
            k+1 by k+2 array is needed.  See details below.

        Each row of the matrix
        """
        self._atom_error_prior[variable_name] = prior_counts.astype("float")
        if len(prior_counts.shape) != 2:
            raise Exception("Expected a 2-d array.")

        row_totals = np.sum(prior_counts > 0, axis=1)
        if not np.all(row_totals > 0):
            raise Exception("Each row must have at least one positive entry")

    def set_level_prior(self,
                        variable_name: str,
                        prior_counts: np.ndarray):
        """
        Set a constrained Dirichlet (0 probabilities are allowed) prior
        distribution on the true levels of a categorical variable.

        Args:
          variable_name: The name of the categorical variable described by the
            prior.
          prior_counts: An array of "prior counts".  Non-positive entries
            indicate prior certainty that the corresponding element is zero.
            The array must have dimension matching the number of levels in the
            variable, and at least one entry must be positive.
        """
        self._level_prior[variable_name] = prior_counts.astype(float).ravel()
        if (np.sum(prior_counts > 0) == 0):
            raise Exception("At least one entry must be positive.")

    def set_level_observation_prior(self,
                                    variable_name: str,
                                    prior_counts: np.ndarray):
        """
        Set a collection of independent constrained (0 probabilities are
        allowed) Dirichlet priors on the conditional probability of the
        observed value given true level.

        Args:
          variable_name: The name of the categorical variable described by the
            prior.
          prior_counts: A 2-d array of "prior counts".  Non-positive entries
            indicate prior certainty that the corresponding element is zero.
            Rows correspond to true level values, and columns correspond to
            actual observations.  The number of rows must therefore match the
            number of levels in the variable.  There is one addtional column
            corresponding to "missing".  Each row of the matrix is treated
            independently, and each row must have at least one positive
            element.
        """
        self._level_observation_prior[variable_name] = (
            prior_counts.astype("float")
        )
        if len(prior_counts.shape) != 2:
            raise Exception("A 2-d array is needed.")
        if prior_counts.shape[1] != prior_counts.shape[0] + 1:
            raise Exception("The matrix needs one more column than rows.")
        positive = prior_counts > 0
        row_ok = positive.sum(axis=1)
        if not np.all(row_ok):
            raise Exception("Each row must have at least one positive element.")

    def impute_rows(self, data, iterations):
        """
        Args:
          data:  Data frame containing the rows to impute.
          iterations:  An array-like collection of iteration numbers to use.

        Returns:
          imputed: A list of imputed data frames.  There is one entry for each
            element in 'iterations'
        """
        imputed = []

        data_table = to_data_table(data)

        for it in iterations:
            self._restore_parameters(it)
            imputed_data_table = self._model.impute_data_set(data_table, 10)
            imputed.append(
                to_data_frame(imputed_data_table, columns=data.columns,
                              index=data.index)
            )
        return imputed

    def plot_row_distribution(self, imputations, row, logscale=True, xlim=None):

        imp = pd.concat([frame.iloc[row, :] for frame in imputations])
        numeric_frame = imp.loc[:, self._numeric_colnames]
        if logscale and self._numeric_colnames:
            numeric_frame = np.log1p(numeric_frame)

        fig, ax = plt.subplots(1, figsize=(10, 8))
        g = sns.boxplot(y="variable", x="value", data=numeric_frame, orient="h")
        xlab = "Value"
        if logscale:
            xlab += " (log1p scale)"
        g.set_xlabel(xlab)
        g.set_title(f"Imputation distribution for row {row+1}.")
        if xlim is not None:
            g.set_xlim(xlim)

    def _allocate_space(self, niter: int):
        xdim = self._model.xdim
        ydim = self._model.ydim
        nclusters = self.nclusters

        self.coefficients = np.empty((niter, xdim, ydim))
        self.residual_variance = np.empty((niter, ydim, ydim))

        self.mixing_distribution = np.empty((niter, nclusters))

        self.atom_probs = {}
        self.atom_error_probs = {}
        self.level_probs = {}
        self.level_observation_probs = {}

        for vname in self._numeric_colnames:
            atom_dim = len(self._atoms[vname])
            self.atom_probs[vname] = np.empty((niter, nclusters, atom_dim + 1))
            self.atom_error_probs[vname] = np.empty(
                (niter, nclusters, atom_dim + 1, atom_dim + 2))

        for vname in self._categorical_colnames:
            nlevels = len(self._levels[vname])
            self.level_probs[vname] = np.empty((niter, nclusters, nlevels))
            self.level_observation_probs[vname] = np.empty(
                (niter, nclusters, nlevels, nlevels + 1))

    def _record_draws(self, iteration: int):
        self.coefficients[iteration, :, :] = self._model.coefficients.to_numpy()
        self.residual_variance[iteration, :, :] = (
            self._model.residual_variance.to_numpy()
        )
        self.mixing_distribution[iteration, :] = (
            self._model.mixing_weights.to_numpy()
        )

        for i in range(len(self._numeric_colnames)):
            vname = self._numeric_colnames[i]
            for cluster in range(self.nclusters):
                self.atom_probs[vname][iteration, cluster, :] = (
                    self._model.atom_probs(cluster, i).to_numpy()
                )
                self.atom_error_probs[vname][iteration, cluster, :, :] = (
                    self._model.atom_error_probs(cluster, i).to_numpy()
                )

        it = iteration
        for i in range(len(self._categorical_colnames)):
            vname = self._categorical_colnames[i]
            for cluster in range(self.nclusters):
                self.level_probs[vname][it, cluster, :] = (
                    self._model.level_probs(cluster, i).to_numpy()
                )
                self.level_observation_probs[vname][it, cluster, :, :] = (
                    self._model.level_observation_probs(cluster, i).to_numpy()
                )

    def _restore_parameters(self, iteration: int):
        """
        Restore the state of the model to a specific MCMC iteration.
        """
        self._model.set_coefficients(
            boom.Matrix(self.coefficients[iteration, :, :]))
        self._model.set_residual_variance(
            boom.SpdMatrix(self.residual_variance[iteration, :, :]))
        for cluster in range(self.nclusters):
            for col in range(len(self._numeric_colnames)):
                vname = self._numeric_colnames[col]
                self._model.set_atom_probs(
                    cluster, col,
                    boom.Vector(self.atom_probs[vname][iteration, cluster, :]))
                self._model.set_atom_error_probs(
                    cluster, col,
                    boom.Matrix(self.atom_error_probs[vname][
                        iteration, cluster, :, :]))

            for col in range(len(self._categorical_colnames)):
                vname = self._categorical_colnames[col]
                self._model.set_level_probs(
                    cluster, col,
                    boom.Vector(self.level_probs[vname][iteration, cluster, :]))
                self._model.set_level_observation_probs(
                    cluster, col,
                    boom.Matrix(self.level_observation_probs[vname][
                        iteration, cluster, :, :]))

    def _set_default_regression_prior(self):
        xdim = self._model.xdim
        ydim = self._model.ydim
        b0 = np.zeros((xdim, ydim))
        Sigma = np.diag(np.ones(ydim))
        self._model.set_regression_prior(
            boom.Matrix(b0),
            1.0,
            boom.SpdMatrix(Sigma),
            ydim + 1
        )

    def _set_default_prior_for_mixing_weights(self):
        counts = np.ones(self.nclusters)
        self._model.set_mixing_weight_prior(boom.Vector(counts))

    @staticmethod
    def _default_atom_prior(atoms):
        """
        The default prior on which atom is the 'truth' places 90% probability on
        the continuous atom and splits the remainder equally on the discrete
        atoms.  An atom that is a string of 9's is given prior weight 0.
        """
        number_of_atoms = len(atoms)
        ans = np.ones(number_of_atoms + 1)
        if number_of_atoms > 0:
            ans *= 0.1 / number_of_atoms
            ans[-1] = 0.9
        for i, atom in enumerate(atoms):
            atom_int = str(int(atom))
            if atom_int.startswith("999") and atom_int.endswith("999"):
                ans[i] = -1
        return ans * 100

    @staticmethod
    def _default_atom_error_prior(number_of_atoms):
        """
        The default prior on the conditional distribution of which atom is
        observed, given which atom is true.

        For discrete atoms the prior is that each atom is either observed
        correctly or marked missing.  For the continuous atom the prior places
        equal weight on all categories.

        """
        ans = np.zeros((number_of_atoms + 1, number_of_atoms + 2))
        for k in range(number_of_atoms):
            ans[k, :] = -1.0
            ans[k, k] = 1.0
            ans[k, -1] = 1.0

        ans[number_of_atoms, :] = 1.0
        return ans

    @staticmethod
    def _default_level_prior(levels):
        """
        The default prior is that all levels are equally likely, except for
        all-blank levels which have prior probability 0.

        Args:
          levels:  Array-like list of unique level values.

        Returns:
          A numpy vector of prior counts.
        """
        nlevels = len(levels)
        string_levels = pd.Series(levels, dtype="str").values
        counts = np.ones(nlevels)
        all_blanks = np.array([s.isspace() for s in string_levels])
        counts[all_blanks] = -1
        return counts

    @staticmethod
    def _default_level_observation_prior(levels):
        """
        The default prior splits prior probability equally between the current
        label and the implicit "missing" label in the final column.
        """
        nlevels = len(levels)
        ans = -1 * np.ones((nlevels, nlevels + 1))
        for i in range(nlevels):
            ans[i, i] = 1.0
            ans[i, nlevels] = 1.0
        return ans

    def _set_prior(self):
        """
        Set the prior distribution on the BOOM model object.  If user-specified
        priors have been set using set_atom_prior, set_atom_error_prior, etc
        then those priors will be installed.  Variables for which no prior was
        specified will receive default priors.
        """
        self._set_default_regression_prior()
        self._set_default_prior_for_mixing_weights()

        for i in range(len(self._numeric_colnames)):
            vname = self._numeric_colnames[i]
            if vname not in self._atom_prior:
                self._atom_prior[vname] = self._default_atom_prior(
                    self._atoms[vname])
            self._model.set_atom_prior(boom.Vector(self._atom_prior[vname]), i)

            if vname not in self._atom_error_prior:
                self._atom_error_prior[vname] = self._default_atom_error_prior(
                    len(self._atoms[vname])
                )
            self._model.set_atom_error_prior(boom.Matrix(
                self._atom_error_prior[vname]), i)

        for i in range(len(self._categorical_colnames)):
            vname = self._categorical_colnames[i]
            levels = self._levels[vname]
            if vname not in self._level_prior:
                self._level_prior[vname] = self._default_level_prior(levels)
            self._model.set_level_prior(boom.Vector(np.array(
                self._level_prior[vname])), i)

            if vname not in self._level_observation_prior:
                self._level_observation_prior[vname] = (
                    self._default_level_observation_prior(levels)
                )
            self._model.set_level_observation_prior(
                boom.Matrix(self._level_observation_prior[vname]), i
            )



import os

# This file contains utilities used to implement unit tests.  For unit tests
# verifying the correctness of utils.py, see test_utils.py.


def delete_if_present(fname):
    """
    Delete file 'fname' if it exists in the filesystem.

    Args:
      fname:  A string or similar object naming the path to a file.
    """
    if os.path.exists(fname):
        os.remove(fname)



import BayesBoom.boom as boom
import pandas as pd
import numpy as np
import BayesBoom.R as R
from scipy.stats import betabinom


class BetaBinomialMixture:
    """
    A finite mixture of BetaBinomial distributions.

    Example use:

    # Create an empty model object.
    model = BetaBinomialMixture()

    # To fit a 3 component mixture, add three components.
    model.add_component(R.BetaPrior(1.0, 1.0), R.UniformPrior(0.1, 1000.0))
    model.add_component(R.BetaPrior(1.0, 1.0), R.UniformPrior(0.1, 1000.0))
    model.add_component(R.BetaPrior(1.0, 1.0), R.UniformPrior(0.1, 1000.0))

    # Add the data.
    model.add_data(my_three_column_matrix)

    # Fit the model by running MCMC for 'niter' iterations.
    model.mcmc(niter=1000)

    """

    def __init__(self, use_data_augmentation=False):
        self._components = list()
        self._mixing_distribution_prior_counts = list()
        self._boom_model = None
        self._use_data_augmentation = use_data_augmentation

    def add_component(self,
                      mean_prior: R.BetaPrior,
                      sample_size_prior: R.DoubleModel,
                      prior_count: float = 1.0):
        """
        Add a beta-binomial mixture component to the model.  The beta-binomial
        distribution is parameterized by two parameters: a and b, which can be
        thought of as counts of successes (a) and failures (b) in a binomial
        inference problem.  The mean of the distribution is a/(a+b), and the
        variance is controlled by the "sample size" a+b.

        Args:
          mean_prior: A prior distribution on the mean (a/a+b) of the component.
          sample_size_prior: A prior distribution on the sample size (a+b) of
            the component.
          prior_count: The mixing distribution has a dirichlet prior, which is
            parameterized by a vector of positive real numbers interpretable as
            prior counts.  This argument is the prior count for the frequency
            of this component in the mixture.
        """
        self._mixing_distribution_prior_counts.append(prior_count)
        self._mixing_weights = None
        component = {
            "mean_prior": mean_prior,
            "sample_size_prior": sample_size_prior
        }
        self._components.append(component)

    def add_data(self, data):
        """
        data: A three-column matrix of numbers giving the (a) number of trials,
          (b) the number of successes, and (c) the number of cases with that
          many trials and successes.

        Example:
        10, 4, 28   # There were 28 cases of 10 trials showing 4 successes.
         5, 0, 2    # There were 2 cases of 5 trials showing no successes.
        """
        if self._boom_model is None:
            self._boom_model = self._build_boom_model()
        if isinstance(data, pd.DataFrame):
            data = data.values
        self._boom_model.add_data(R.to_boom_matrix(data.astype(float)))

    def mcmc(self, niter, ping=None, seed=None):
        """
        Run the MCMC algorithm for 'niter' iterations.

        Args:
          niter:  The desired number of MCMC iterations.
          ping:  The frequency with which to print status updates.
          seed:  The random seed to use for the C++ random number generator.

        Returns:
          None

        Effects:
          MCMC draws are stored in the object.
        """
        if self._boom_model is None:
            self._boom_model = self._build_boom_model()
            import warnings
            warnings.warn("Running MCMC on a model with no data assigned.")

        self._create_storage(niter)
        if seed is not None:
            # I don't have a lot of confidence that this works.
            boom.GlobalRng.rng.seed(int(seed))
        for i in range(niter):
            R.report_progress(i, ping)
            self._boom_model.sample_posterior()
            self._record_draw(i)

    @property
    def number_of_mixture_components(self):
        """
        The number of mixture components contained in the object.
        """
        return len(self._mixing_distribution_prior_counts)

    @property
    def niter(self):
        """
        The number of MCMC iterations that have been run.
        """
        if self._mixing_weights is None:
            return 0
        return self._mixing_weights.shape[0]

    @property
    def a(self):
        """
        MCMC draws of the 'success count' parameters.  This is a matrix with
        'niter' rows and 'number_of_mixture_components' columns, with each
        column representing the 'a' parameter of one mixture component.
        """
        ans = np.empty((self.niter, self.number_of_mixture_components))
        for i in range(self.number_of_mixture_components):
            ans[:, i] = self._components[i]["draws"][:, 0]
        return ans

    @property
    def b(self):
        """
        MCMC draws of the 'failure count' parameters.  This is a matrix with
        'niter' rows and 'number_of_mixture_components' columns, with each
        column representing the 'b' parameter of one mixture component.
        """
        ans = np.empty((self.niter, self.number_of_mixture_components))
        for i in range(self.number_of_mixture_components):
            ans[:, i] = self._components[i]["draws"][:, 1]
        return ans

    @property
    def means(self):
        """
        MCMC draws of the mean parameter for each mixture component.
        """
        a = self.a
        b = self.b
        return a / (a + b)

    @property
    def sample_sizes(self):
        """
        MCMC draws of the sample size parameter for each mixture component.
        """
        return self.a + self.b

    @property
    def mixing_weights(self):
        """
        MCMC draws of the mixing weights.
        """
        return self._mixing_weights

    def cluster_membership_probabilities(
            self, trials: int, successes: int, burn: int = 0):
        """
        Return the cluster membership probabilities for a user with the given
        number of successes and trials.

        Args:
          trials:  A number of binomial trials.
          successes:  A success count 0 <= successes <= trials.
          burn:  The number of MCMC iterations to discard as burn-in.

        Returns:
          A vector of cluster membership probabilities.
        """
        from .utils import normalize_logprob

        log_probs = np.log(self._mixing_weights)
        a = self.a
        b = self.b
        if burn is not None and burn > 0:
            a = a[burn:, :]
            b = b[burn:, :]
            log_probs = log_probs[burn:, :]

        for component in range(log_probs.shape[1]):
            log_probs[:, component] += betabinom.logpmf(
                successes, trials, a[:, component], b[:, component])

        probs = normalize_logprob(log_probs)
        return probs.mean(axis=0)

    def _create_storage(self, niter: int):
        """
        Create storage for 'niter' MCMC draws.
        """
        for component in self._components:
            component["draws"] = np.empty((niter, 2))
        self._mixing_weights = np.empty(
            (niter, self.number_of_mixture_components))

    def _record_draw(self, i):
        for c, component in enumerate(self._components):
            component["draws"][i, :] = [
                self._boom_model.mixture_component(c).a,
                self._boom_model.mixture_component(c).b
            ]
        self._mixing_weights[i, :] = (
            self._boom_model.mixing_distribution.probs.to_numpy()
        )

    def _build_boom_model(self):
        """
        Build a boom_model from
        """
        prior_counts = np.array(self._mixing_distribution_prior_counts)
        prior_mixing_distribution = prior_counts / np.sum(prior_counts)

        mixing_distribution = boom.MultinomialModel(
            R.to_boom_vector(prior_mixing_distribution))
        dirichlet_prior = boom.DirichletModel(R.to_boom_vector(prior_counts))

        mixing_distribution_sampler = boom.MultinomialDirichletSampler(
            mixing_distribution, dirichlet_prior)
        mixing_distribution.set_method(mixing_distribution_sampler)

        boom_components = list()
        mean_priors = list()
        sample_size_priors = list()
        for component in self._components:
            mean_prior = component["mean_prior"]
            sample_size_prior = component["sample_size_prior"]
            initial_a = mean_prior.mean * sample_size_prior.mean
            initial_b = sample_size_prior.mean - initial_a
            component_model = boom.BetaBinomialModel(initial_a, initial_b)
            if self._use_data_augmentation:
                component_sampler = boom.BetaBinomialPosteriorSampler(
                    component_model,
                    mean_prior.boom(),
                    sample_size_prior.boom())
                component_model.set_method(component_sampler)
            else:
                mean_priors.append(mean_prior.boom())
                sample_size_priors.append(sample_size_prior.boom())

            boom_components.append(component_model)

        boom_model = boom.BetaBinomialMixtureModel(
            boom_components, mixing_distribution)
        if self._use_data_augmentation:
            sampler = boom.BetaBinomialMixturePosteriorSampler(boom_model)
        else:
            sampler = boom.BetaBinomialMixtureDirectPosteriorSampler(
                boom_model, dirichlet_prior, mean_priors, sample_size_priors)
        boom_model.set_method(sampler)
        return boom_model



import BayesBoom.boom as boom
import numpy as np
import BayesBoom.R as R
from .state_models import StateModel
import patsy


class DynamicRegressionStateModel(StateModel):
    """
    The state in this model represents a vector of regression coefficients that
    evolves over time.  The state equation is thus

       y[t] = beta[t].dot(x[t]) + other_state[t] + error[t]

    The variables x[t] are determined by a model formula and a data set
    supplied at contruction time.

    The state transition equation is a random walk.

       beta[t+1] = beta[t] + error

    The vector of errors is independent zero-mean Gaussian noise, where
    component i has standard deviation sigma[i].
    """

    def __init__(self,
                 formula,
                 data,
                 sigma_prior=None,
                 initial_state_prior=None):
        """
        Args:

          formula: A string describing a regression relationship in the R
            modeling language.  For example "y ~ x1 + x3" or "y ~ x1 * x2 + x3".
            The 'dot' operator is not understood, but "y ~ " + dot(
        """
        self._formula = formula + "- 1"
        response, self._predictors = patsy.dmatrices(self._formula, data=data)
        xtx = self._predictors.T @ self._predictors
        xty = self._predictors.T @ response
        sdy = np.nanstd(response)
        sdx = np.nanstd(self._predictors, axis=0)
        self._sigma_prior = self._verify_prior(sigma_prior, sdy, sdx)
        self._initial_state_prior = self._verify_initial_state_prior(
            initial_state_prior, xtx, xty, sdy)
        self._build_state_model()

    def __repr__(self):
        ans = f"DynamicRegressionStateModel: {self._formula}"
        return ans

    @property
    def label(self):
        if len(self._formula) > 40:
            return "DynReg(long formula)"
        else:
            return f"DynReg({self._formula})"

    @property
    def state_dimension(self):
        return self._predictors.shape[1]

    @property
    def xdim(self):
        return self.state_dimension

    def allocate_space(self, niter, time_dimension):
        nrows = self._predictors.shape[0]
        if time_dimension != nrows:
            raise Exception(f"time_dimension = {time_dimension}, but "
                            f"self._predictors has {nrows} rows.")
        self.sigma_draws = np.zeros((niter, self.xdim))
        self._state_contribution = np.zeros((niter, time_dimension))
        self._dynamic_coefficients = np.zeros((
            niter, self.xdim, time_dimension))

    def record_state(self, iteration, state_matrix):
        if self._state_index < 0:
            raise Exception("Each state model must be told where its state"
                            "component begins in the global state vector.  "
                            "Try calling set_state_index.")
        self.sigma_draws[iteration, :] = self._state_model.sigma.to_numpy()

        begin = self._state_index
        end = self._state_index + self.state_dimension
        self._dynamic_coefficients[iteration, :, :] = state_matrix[begin:end, :]
        self._state_contribution[iteration, :] = (
            self._dynamic_coefficients[iteration, :, :] * self._predictors.T
            ).sum(axis=0)

    def restore_state(self, iteration):
        self._state_model.set_sigma(self.sigma_draws[iteration, :])

    @property
    def state_contribution(self):
        return self._state_contribution

    def __getstate__(self):
        payload = self.__dict__.copy()
        del payload["_state_model"]
        return payload

    def __setstate__(self, payload):
        self.__dict__ = payload
        self._build_state_model()

    def _validate_initial_distributions(self, initial_state_prior, y, sdy):
        if initial_state_prior is None:
            if sdy is None:
                sdy = np.nanstd(y, ddof=1)
            initial_state_prior = R.NormalPrior(0.0, float(sdy))
        if not isinstance(initial_state_prior, R.NormalPrior):
            raise Exception(
                "initial_state_prior should be an R.NormalPrior.")
        self._initial_state_prior = initial_state_prior

    def _build_state_model(self):
        self._state_model = boom.DynamicRegressionStateModel(
            R.to_boom_matrix(self._predictors))

        boom_sigma_priors = [pri.boom() for pri in self._sigma_prior]
        state_model_sampler = boom.DynamicRegressionIndependentPosteriorSampler(
            self._state_model, boom_sigma_priors)
        for i, prior in enumerate(self._sigma_prior):
            finite_limit = np.isfinite(prior.upper_limit)
            if prior.upper_limit > 0 and finite_limit:
                state_model_sampler.set_sigma_max(i, prior.upper_limit)
        self._state_model.set_method(state_model_sampler)

        self._state_model.set_initial_state_mean(
            R.to_boom_vector(self._initial_state_prior.mean))
        self._state_model.set_initial_state_variance(
            R.to_boom_spd(self._initial_state_prior.Sigma))

    def _verify_prior(self, sigma_prior, sdy, sdx):
        if sigma_prior is None:
            self._sigma_prior = [
                R.SdPrior(.01 * sdy / sdxi, 1) for sdxi in sdx]
        elif isinstance(sigma_prior, R.SdPrior):
            self._sigma_prior = [sigma_prior] * len(sdx)

        if not R.is_iterable(self._sigma_prior) and all(
            [isinstance(x, R.SdPrior) for x in self._sigma_prior]
        ):
            raise Exception(
                "sigma_prior must be a list-like of R.SdPrior objects."
            )

        return self._sigma_prior

    def _verify_initial_state_prior(
            self, initial_state_prior, xtx, xty, sdy):
        if initial_state_prior is None:
            try:
                beta_hat = np.linalg.solve(xtx, xty)
                if not np.all(np.finite(beta_hat)):
                    raise Exception("Least squares initializer failed.")

                self._initial_state_prior = R.MvnPrior(
                    beta_hat, sdy * sdy * np.linalg.inv(xtx))
            except Exception:
                self._initial_state_prior = R.MvnPrior(
                    np.zeros(self.xdim),
                    sdy * sdy * np.diag(1.0 / np.diagonal(xtx)))

        elif isinstance(initial_state_prior, R.NormalPrior):
            mean = np.full(initial_state_prior.mean, self.xdim)
            var = np.full(initial_state_prior.sd ** 2, self.xdim)
            self._initial_state_prior = R.MvnPrior(mean, np.diag(var))

        elif isinstance(initial_state_prior, list) and all(
                [isinstance(x, R.NormalPrior) for x in initial_state_prior]):
            mean = np.array([x.mean for x in initial_state_prior])
            var = np.array([x.sd ** 2 for x in initial_state_prior])
            self._initial_state_prior = R.MvnPrior(mean, np.diag(var))

        else:
            if not isinstance(initial_state_prior, R.MvnPrior):
                raise Exception("Unrecognized type for initial_state_prior.")
            self._initial_state_prior = initial_state_prior

        return self._initial_state_prior



import unittest
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

import BayesBoom.R as R

from BayesBoom.R.plots import (
    mosaic_plot,
    plot_dynamic_distribution,
    time_series_boxplot,
    hosmer_lemeshow_plot,
    pretty_plot_ticks,
)


def char_range(c1, c2):
    """Generates the characters from `c1` to `c2`, inclusive."""
    for c in range(ord(c1), ord(c2) + 1):
        yield chr(c)


LETTERS = list(char_range("a", "z"))

us_states = ["AL", "AK", "AZ", "AR", "CA", "CO", "CT", "DC", "DE", "FL", "GA",
             "HI", "ID", "IL", "IN", "IA", "KS", "KY", "LA", "ME", "MD", "MA",
             "MI", "MN", "MS", "MO", "MT", "NE", "NV", "NH", "NJ", "NM", "NY",
             "NC", "ND", "OH", "OK", "OR", "PA", "RI", "SC", "SD", "TN", "TX",
             "UT", "VT", "VA", "WA", "WV", "WI", "WY"]


class TestPlots(unittest.TestCase):

    def setUp(self):
        pass

    def test_plot_points(self):
        x = np.sort(np.random.randn(100))
        y = 3 + .4 * x + np.random.randn(100) * .2
        fig, ax = plt.subplots()
        ax.scatter(x, y)
        if _show_figs:
            fig.show()

    def test_fill_between(self):
        x = np.sort(np.random.randn(100))
        y = np.abs(3 + .4 * x + np.random.randn(100) * .2)
        negy = -1 * y
        fig, ax = plt.subplots()
        ax.fill_between(x, y, negy, color="black", edgecolor="none")
        if _show_figs:
            fig.show()

    def test_plot_dynamic_distribution(self):
        time_dimension = 50
        niter = 500
        noise = np.random.randn(niter, time_dimension)
        random_walks = np.cumsum(noise, axis=1)

        timestamps = np.arange(0, time_dimension)
        fig, ax = plt.subplots()

        plot_dynamic_distribution(curves=random_walks, timestamps=timestamps,
                                  ax=ax, quantile_step=.01)
        if _show_figs:
            fig.show()

    def test_mosaic_plot(self):
        counts = np.random.randint(1, 10, (12, 4))
        counts = pd.DataFrame(counts, index=us_states[:counts.shape[0]],
                              columns=["red", "blue", "green", "yellow"])

        print(counts)
        fig, ax = plt.subplots()
        foo = mosaic_plot(counts, ax=ax)

        if _debug_mode:
            fig.show()

        self.assertIsInstance(foo, plt.Axes)

        if _show_figs:
            fig.show()

    def test_time_series_boxplot(self):
        x = np.random.randn(100, 8)
        time = pd.date_range(start="2020-02-1", periods=8, end="2020-02-08")
        fig, ax = plt.subplots()
        time_series_boxplot(x, time, ax=ax)
        if _show_figs:
            fig.show()

    def test_lty(self):
        x = np.linspace(0, 10)
        fig, ax = plt.subplots()
        for i in range(10):
            y = x + i
            ax.plot(x, y, ls=R.lty(i))
        if _show_figs:
            fig.show()

    def test_hosmer(self):
        x = np.random.uniform(.25, .85, 1000)
        rnd = np.random.uniform(0, 1, 1000)
        y = rnd < x
        fig, ax = plt.subplots()
        foo = hosmer_lemeshow_plot(y, x, main="Hosmer-Lemeshow plot", ax=ax)
        self.assertIsInstance(foo, plt.Axes)
        if _show_figs:
            fig.show()


_debug_mode = False
_show_figs = _debug_mode


if _debug_mode:
    import pdb  # noqa

    # Turn warnings into errors.
    # warnings.simplefilter("error")

    # Run the test you are trying to debug here.  Instantiate the test class,
    # then call the problematic test.  Call pdb.pm() in the event of an
    # exception.
    print("Hello, world!")

    rig = TestPlots()
    if hasattr(rig, "setUpClass"):
        rig.setUpClass()
    if hasattr(rig, "setUp"):
        rig.setUp()

    # rig.test_plot_points()
    # rig.test_mosaic_plot()
    rig.test_hosmer()
    # rig.test_lty()
    # rig.test_fill_between()

    print("Goodbye, cruel world!")

else:
    if __name__ == "__main__":
        unittest.main(verbosity=2)



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from pandas.api.types import is_numeric_dtype

from numbers import Number

from .R import (
    data_range,
    remove_common_suffix,
    remove_common_prefix,
    unique_match
)


# ===========================================================================
# Graphics related utilities.  These do not interact with a figure, axes, or
# graphics device.
# ===========================================================================
def _skim_plot_options(xlab="", ylab="", xlim=None, ylim=None, title="",
                       main="", **kwargs):
    """
    Remove plotting options used by the R package from other options supplied by
    kwargs.  This allows the remaining options to be later passed to pyplot.

    Returns:
      plot_options:  The options expected by R plotting functions.a
      kwargs:  Everything not pulled into plot_options.
    """
    plot_options = {'xlab': xlab,
                    'ylab': ylab,
                    'xlim': xlim,
                    'ylim': ylim,
                    'main': main,
                    'title': title}
    return plot_options, kwargs


def _set_plot_options(ax, xlab="", ylab="", xlim=None, ylim=None, title="",
                      main="", **kwargs):
    if xlab != "":
        ax.set_xlabel(xlab)
    if ylab != "":
        ax.set_ylabel(ylab)
    if xlim is not None:
        ax.set_xlim(xlim)
    if ylim is not None:
        ax.set_ylim(ylim)
    if main != "":
        title = main
    if title != "":
        ax.set_title(title)


def pretty_plot_ticks(low, high, n):
    """
    Return a set of 'pretty' tick locations.

    Args:
      low: The lower end of the plotting range.
      high: The upper end of the plotting range.
      n: The desired number of ticks.

    Returns:
      A np.array with 'pretty' values suitable for tick labels.  The length of
      the array will typically not be exactly 'n'.  Its lower endpoint will be
      <= low and its upper endpoint will be >= high.

    Taken from StackOverflow:
    https://stackoverflow.com/questions/43075617/python-function-equivalent-to-rs-pretty
    """
    def nicenumber(x, round_result: bool):
        exp = np.floor(np.log10(x))
        f = x / 10**exp

        if round_result:
            if f < 1.5:
                nf = 1.
            elif f < 3.:
                nf = 2.
            elif f < 7.:
                nf = 5.
            else:
                nf = 10.
        else:
            if f <= 1.:
                nf = 1.
            elif f <= 2.:
                nf = 2.
            elif f <= 5.:
                nf = 5.
            else:
                nf = 10.

        return nf * 10.**exp

    num_range = nicenumber(high - low, False)
    d = nicenumber(num_range / (n - 1), True)
    miny = np.floor(low / d) * d
    maxy = np.ceil(high / d) * d
    return np.arange(miny, maxy+0.5*d, d)


def plot_grid_shape(nplots: int):
    """
    Compute the number of rows and columns needed to plot 'nplots'.

    :param nplots:
        The desired number of plots.

    :return tuple:
        The number of rows, and columns, needed to plot that many plots.
    """
    nr = int(max(1, np.sqrt(nplots)))
    nc = int(np.ceil(nplots / nr))
    return nr, nc


# ===========================================================================
# Low level plot functions.  These interact with the current axes object in the
# current graphics device.  They do not advance to the next axes.
# ===========================================================================
def points(x, y, ax, s=None, **kwargs):
    """
    Add points to the plot showing on the current graphics device.
    """
    if s is None:
        s = 20 / np.sqrt(len(y))
    ax.scatter(x, y, **kwargs)


def lines(x, y, ax, **kwargs):
    """
    Add lines to the most recent plot.
    """
    ax.plot(x, y, **kwargs)


def abline(ax, a=0, b=1, h=None, v=None, **kwargs):
    """
    Add a line with specified slope and intercept to a plot.

    Args:
      a: The intercept of the line.
      b: The slope of the line.
      h: Draw a horizontal line at this y value.
      v: Draw a vertical line at this x value.
      **kwargs:  Additional arguments passed to 'plt.plot'.

    Returns:
      None

    Effect:
      The requested line is plotted on 'ax'.
    """
    xlim = ax.get_xlim()
    ylim = ax.get_ylim()
    if h is not None:
        p0 = xlim[0], h
        p1 = xlim[1], h
    elif v is not None:
        p0 = v, ylim[0]
        p1 = v, ylim[1]
    else:
        p0 = xlim[0], a + b * xlim[0]
        p1 = xlim[1], a + b * xlim[1]

    x = [p0[0], p1[0]]
    y = [p0[1], p1[1]]
    ax.plot(x, y, **kwargs)


# ===========================================================================
# High level plot functions.  These advance to the next active axes object in
# the graphics device.  If the graphics device is already at the last set of
# axes, the plot resets.
# ===========================================================================

def plot(x, y=None, s=None, hexbin_threshold=1e+5, ax=None, **kwargs):
    """
    For now, a 'plot' is a scatterplot.  At some point I will make 'plot'
    generic as with R.
    """
    if ax is None:
        fig, ax = plt.subplots(1, 1)
        draw = True
    else:
        draw = False

    plot_options, kwargs = _skim_plot_options(**kwargs)

    if y is None:
        y = x
        x = np.arange(len(y))

    sample_size = len(x)

    if s is None:
        s = 20 / np.sqrt(sample_size)
    if sample_size < hexbin_threshold:
        ax.scatter(x, y, s=s, **kwargs)
    else:
        ax.hexbin(x, y, **kwargs)
    _set_plot_options(ax, **plot_options)

    if draw:
        fig.show()

    return ax


def hist(x, density: bool = False, edgecolor="black", color=".75", add=False,
         ax=None, **kwargs):
    """
    Plot a histogram of x.

    Args:
      x: The variable to be plotted.
      density: If True then the area of the histogram bars sums to 1.
    """
    fig = None
    if ax is None:
        fig, ax = plt.subplots(1, 1)

    plot_options, kwargs = _skim_plot_options(**kwargs)
    ax.hist(x[np.isfinite(x)], edgecolor=edgecolor, density=density,
            color=color, **kwargs)
    _set_plot_options(ax, **plot_options)

    if fig is not None:
        fig.show()

    return fig, ax


def barplot(x, labels=None, zero=True, ax=None, **kwargs):
    """
    Make a horizonal bar plot.

    Args:
      x:  Array-like collection of numbers to plot.
      labels: Labels for the bars.  If x is a pd.Series then labels==None means
        to take the labels from the series index.  Otherwise None means don't
        plot labels.
      zero:  Bool.  Should zero be forcibly included in the numeric axis.
      xlab: Label for the horizonal "numeric" axis.
      ylab: Label for the vertical "categorial" axis.
      kwargs: extra arguments passed to plt.subplots or plt.barh.

    """
    if ax is None:
        fig, ax = plt.subplots(1, 1)
        draw = True
    else:
        draw = False

    x = x[::-1]
    if labels is not None:
        labels = labels[::-1]

    if labels is None and isinstance(x, pd.Series):
        labels = [str(lab) for lab in x.index]
    bar_locations = np.arange(len(x))
    plot_options, kwargs = _skim_plot_options(**kwargs)

    if labels is None:
        labels = [""] * len(x)

    ax.barh(bar_locations, x, height=.8, **kwargs)
    ax.set_yticks(bar_locations)
    ax.set_yticklabels(labels)
    lo = np.nanmin(x)
    hi = np.nanmax(x)
    if zero and lo > 0:
        lo = 0.0
    if zero and hi < 0:
        hi = 0.0

    ax.set_xticks(pretty_plot_ticks(lo, hi, 5))
    _set_plot_options(ax, **plot_options)

    if draw:
        fig.show()

    return ax


def boxplot(x, labels=None, ax=None, **kwargs):
    if ax is None:
        fig, ax = plt.subplots(1, 1)
        draw = True
    else:
        draw = False

    if labels is None and isinstance(x, pd.DataFrame):
        labels = x.columns

    ax.boxplot(x, vert=False)
    _set_plot_options(ax, **kwargs)

    if draw:
        fig.show()

    return ax


def time_series_boxplot(curves, time=None, ylim=None, ax=None, **kwargs):
    """
    Plot side-by-side boxplots showing the evolution of a distribution over
    time.

    Args:
      curves: A matrix or data frame.  Rows represent different curves or Monte
        Carlo draws.  Columns represent time.

      time: A collectiond of timestamps corresponding to the column labels of
        'curves'.  If None then timestamps will be taken as the column labels
        of 'curves' (if curves is a DataFrame), or else they will be assigned
        the indices 1, 2, 3, ... .

      ylim: Limits on the vertical axis.

      ax:  The plt.Axes object on which to draw the plot.

      **kwargs:  Extra arguments passed to 'boxplot'.
    """
    plot_options, kwargs = _skim_plot_options(**kwargs)
    _set_plot_options(ax, **plot_options)

    if time is None:
        if isinstance(curves, pd.DataFrame):
            time = curves.columns
        else:
            time_dim = curves.shape[1]
            time = np.linspace(
                1, time_dim, num=time_dim).astype(int).astype(str)

    time = remove_common_prefix(
        remove_common_suffix(
            [str(x) for x in time]))
    ax.boxplot(curves, labels=time, **kwargs)
    return ax


def plot_ts(x, timestamps=None, ax=None, **kwargs):
    """ Plot a time series."""

    if ax is None:
        fig, ax = plt.subplots(1, 1)

    if timestamps is None:
        if isinstance(x, pd.Series):
            timestamps = x.index

    ax.plot(x)
    _set_plot_options(ax, **kwargs)

    return ax


def pairs(x, **kwargs):
    """
    Return a pairs plot of the data in x.

    Args:
      x: A pd.DataFrame or numpy matrix.

    Returns:
      The seaborn.axisgrid.PairGrid object containing the plots.
    """
    import seaborn as sns
    if isinstance(x, np.ndarray):
        x = pd.DataFrame(x)
    ans = sns.pairplot(x, **kwargs)
    ans.fig.show()
    return ans


# ===========================================================================
# Custom plots
# ===========================================================================

def mosaic_plot(counts, ax=None, col_vname=None, row_vname=None):
    """
    Args:
      counts: A pd.DataFrame or equivalent, containing the contingency table
        describing the relationship between two categorical variables.
      ax:
    """
    if ax is None:
        fig, ax = plt.subplots(1, 1)
    else:
        fig = None

    if not isinstance(ax, plt.Axes):
        raise Exception("ax must be a plt.Axes object")

    if isinstance(counts, np.ndarray):
        counts = pd.DataFrame(counts,
                              index=np.arange(counts.shape[0]).astype(str),
                              columns=np.arange(counts.shape[1]).astype(str))

    # The margininal distribution of the variable described by the rows,
    # obtained by summing over columns.
    row_margin = np.array(counts.sum(axis=1))
    nrow = len(row_margin)

    # The margininal distribution of the variable described by the columns,
    # obtained by summing over rows.
    col_margin = np.array(counts.sum(axis=0))
    ncol = len(col_margin)

    # The conditional distribution of the row variable, within each column.
    conditional = pd.DataFrame(counts / col_margin, index=counts.index,
                               columns=counts.columns)

    col_margin = col_margin / np.sum(col_margin)
    row_margin = row_margin / np.sum(row_margin)

    cum_col_margin = np.cumsum(col_margin)
    lower_col_margin = np.array([0] + cum_col_margin[:-1].tolist())
    column_positions = (lower_col_margin + cum_col_margin) / 2
    column_widths = cum_col_margin - lower_col_margin

    lower = np.zeros(ncol)
    for row_index in range(len(row_margin)):
        ax.bar(column_positions, conditional.iloc[row_index, :],
               width=column_widths, bottom=lower, edgecolor="gray",
               label=counts.index[row_index])
        lower += conditional.iloc[row_index, :]

    ax.set_xticks(column_positions)
    ax.set_xticklabels(counts.columns)

    cum_row_margin = np.cumsum(row_margin)
    row_low = np.array([0] + cum_row_margin[:-1].tolist())

    if np.min(row_margin) < .05:
        ax.set_yticks(np.linspace(0, 1, nrow))
    else:
        row_tick_locations = (cum_row_margin + row_low) / 2
        ax.set_yticks(row_tick_locations)
    ax.set_yticklabels(counts.index)

    if fig is not None:
        fig.show()

    return ax


def histabunch(data, min_continuous=12, max_levels=40, same_scale=False):
    nvars = data.shape[1]
    nr, nc = plot_grid_shape(nvars)
    _, ax = plt.subplots(nr, nc)

    def is_all_missing(y):
        return y.count() == 0

    def is_numeric(y):
        if isinstance(y, np.ndarray):
            y = pd.Series(y)
        return is_numeric_dtype(y) and len(y.value_counts()) >= min_continuous

    def is_categorical(y):
        return not is_numeric(y)

    def hist_categorical(y, vname, max_levels, ax):
        if isinstance(np.ndarray(y)):
            y = pd.Series(y)

        counts = y.value_counts()
        if counts.shape[0] > max_levels:
            counts = counts.sort_values(ascending=False)
            counts = counts[:max_levels]
        barplot(counts, ax=ax, main=vname)

    def hist_numeric(y, vname, ax):
        hist(y, ax=ax, main=vname)

    def plot_all_missing(vname, ax):
        pass

    vnames = data.columns
    plot_number = 0
    for row in range(nr):
        for col in range(nc):
            # Life would be better if we could reshape ax to ensure that it was
            # a matrix.  But alas.
            if len(ax.shape) == 2:
                ax_index = (row, col)
            else:
                ax_index = row + col
            if plot_number < nvars:
                vname = vnames[plot_number]
                y = data.loc[:, vname]

                if is_all_missing(y):
                    plot_all_missing(vname, ax[ax_index])
                elif is_numeric(y):
                    hist_numeric(y, vname, ax[ax_index])
                elif is_categorical(y):
                    hist_categorical(y, vname, max_levels, ax=ax[ax_index])
                plot_number += 1


def plot_many_ts(series, same_scale=True, ylim=None, gap=0, truth=None,
                 **kwargs):
    """
    Args:

      series: A numpy array of data to be plotted.  The first dimension of the
        array is time.
      same_scale:  If True then all series are plotted on the same scale.
      ylim:  lower and upper limits of the Y axis.
      gap:  Amount of space to leave beween panels of the plot.
      truth: Values at which reference lines should be drawn.  Either a single
        numeric value, or an array of values that matches the shape of 'series'
        (without the time dimension).
      **kwargs:  Other keyword arguments are ignored.
    """
    series = np.array(series)

    if len(series.shape) == 2:
        nseries = series.shape[1]
        nr, nc = plot_grid_shape(nseries)
    elif len(series.shape) == 3:
        nr, nc = series.shape[1], series.shape[2]
        nseries = nr * nc
    else:
        raise Exception("Wrong shape for 'series' argument to plot_many_ts.")

    gap *= .2
    fig, ax = plt.subplots(nr, nc)
    fig.subplots_adjust(hspace=gap, wspace=gap)

    series_number = 0
    if same_scale:
        ylim = data_range(series)

    if truth is not None:
        if isinstance(truth, Number):
            truth = np.ones((nr, nc)) * truth
        elif len(truth.shape) == 1:
            truth = np.concatenate(
                truth, np.zeros(nr * nc - nseries)).reshape((nr, nc))

    for i in range(nr):
        for j in range(nc):
            ax[i, j].set_frame_on(True)
            ax[i, j].xaxis.set_tick_params(bottom=False, top=False,
                                           labelbottom=False, labeltop=False)
            ax[i, j].yaxis.set_tick_params(left=False, right=False,
                                           labelleft=False, labelright=False)

            # In the top and bottom row, set x axis to be either the bottom or
            # the top.
            if i == 0:
                if (j % 2) == 1:
                    ax[i, j].xaxis.set_tick_params(
                        top=True, labeltop=True)

            if i + 1 == nr:
                if (j % 2) == 0:
                    ax[i, j].xaxis.set_tick_params(
                        bottom=True, labelbottom=True)

            if same_scale:
                # In the left or right column, set the y axis to be either the
                # left or the right.
                ax[i, j].set_ylim(ylim)
                if j == 0 and (i % 2) == 0:
                    ax[i, j].yaxis.set_tick_params(left=True, labelleft=True)
                if j + 1 == nc and (i % 2) == 1:
                    ax[i, j].yaxis.set_tick_params(right=True, labelright=True)
                    ax[i, j].yaxis.tick_right()

            else:
                if ylim is not None:
                    print("setting ylim")
                    ax[i, j].set_ylim(ylim)

            if series_number < nseries:
                if len(series.shape) == 2:
                    the_series = series[:, series_number]
                else:
                    the_series = series[:, i, j]
                ax[i, j].plot(the_series)
                if truth is not None:
                    abline(ax=ax[i, j], h=truth[i, j])
                series_number += 1

    return fig, ax


def plot_dynamic_distribution(
        curves,
        timestamps=None,
        quantile_step=.02,
        xlab="Time",
        ylab="distribution",
        col="black",
        highlight_median="",
        ax=None,
        **kwargs):
    """
    Plot a dynamic distribution represented by a collection of curves simulated
    from that distribution.

    Args:
      curves:
        A numpy matrix of time series.  Rows correspond to different series,
        and columns correspond to time.
      timestamps:
        An array-like collection of increasing time stamps corresponding to the
        time points in 'curves.'
      quantile_step:
        The plotted distribution is formed by taking the quantiles of the
        curves at each time point.  The smaller the value of quantile_step the
        finer the approximation, but the larger and slower the plot.
      xlim:
        The limits on the horizontal axis.
      xlab:
        The label for the horizontal axis.
      highlight_median:
        The name of the color used to highlight the median.  The empty string
        means not to add extra highlighting.

      **kwargs: Extra arguments passed to _skim_plot_options.
    """
    redraw = False
    if ax is None:
        # device = get_current_graphics_device()
        # ax = device.next_axes
        fig, ax = plt.subplots(1, 1)
        redraw = True

    plot_options, kwargs = _skim_plot_options(**kwargs)

    quantile_points = np.arange(0, 1, quantile_step)
    curve_quantiles = np.quantile(curves, q=quantile_points, axis=0)
    if timestamps is None:
        timestamps = np.arange(curves.shape[1])
    if len(timestamps) != curves.shape[1]:
        raise Exception("len(timestamps) must match curves.shape[1].")

    for i in range(int(np.floor(len(quantile_points) / 2))):
        lo = curve_quantiles[i, :]
        hi = curve_quantiles[-1-i, :]
        ax.fill_between(timestamps, lo, hi, color=col, edgecolor="none",
                        alpha=(i / len(quantile_points)))

    if highlight_median != "":
        ax.plot(timestamps, np.nanmedian(curves, axis=0),
                color=highlight_median)

    _set_plot_options(ax, **plot_options)

    if redraw:
        # device = get_current_graphics_device()
        # device.draw_current_axes()
        fig.show()

    return ax


# Allow the same function name as R.
PlotDynamicDistribution = plot_dynamic_distribution


def compare_dynamic_distributions(
        list_of_curves,
        timestamps,
        style="dynamic",
        xlab="Time",
        ylab="",
        frame_labels=None,
        main="",
        actuals=None,
        col_actuals=None,
        pch_actuals="o",
        cex_actuals=1,
        vertical_cuts=None,
        fig=None,
        **kwargs):
    """
    Produce a plot showing several stacked dynamic distributions over the same
    horizontal axis.

    Args:
      list.of.curves: A list of matrices, all having the same number of
         columns.  Each matrix represents a distribution of curves, with rows
         corresponding to individual curves, and columns to time points.
      timestamps: A vector of time stamps, with length matching the number of
        columns in each element of list.of.curves.
      style: Should the curves be represented using a dynamic distribution
        plot, or boxplots.  Boxplots are better for small numbers of time
        points.  Dynamic distribution plots are better for large numbers of
        time points.
      xlab:  Label for the horizontal axis.
      ylab:  Label for the (outer) vertical axis.
      frame.labels: Labels for the vertical axis of each subplot. The length
        must match the number of plot.
      main:  Main title for the plot.
      actuals: If non-NULL, actuals should be a numeric vector giving the
        actual "true" value at each time point.
      col_actuals:  Color to use for the actuals.
      cex_actuals:  Scale factor for actuals.
      vertical.cuts: If not None then this must be a vector of the same type as
        'timestamps' with length matching the number of plots.  A vertical line
        will be drawn at this location for each plot.  Entries with the value
        NaN or NaT signal that no vertical line should be drawn for that entry.
      kwargs: Extra arguments passed to PlotDynamicDistribution or
       TimeSeriesBoxplot.
    """
    style = unique_match(style, ["dynamic", "boxplot"])
    nplots = len(list_of_curves)
    ntimes = len(timestamps)
    for i in range(nplots):
        if list_of_curves[i].shape[1] != ntimes:
            raise Exception(f"Entry {i} in 'list_of_curves' did not have the "
                            f"right number of columns.  Expected {ntimes}, "
                            f"got {list_of_curves[i].shape[1]}.")
    if frame_labels is None:
        frame_labels = [str(i + 1) for i in range(nplots)]
    if not len(frame_labels) == nplots:
        raise Exception("frame labels do not match number of curves.")
    if fig is None:
        fig = plt.figure()

    ax = fig.subplots(nplots, 1, sharex=True)
    if nplots == 1:
        # If there is only one plot, put ax in a list so we can "iterate" over
        # it without breaking the multiplot code.
        ax = [ax]
    for i in range(nplots):
        if style == "dynamic":
            plot_dynamic_distribution(list_of_curves[i],
                                      timestamps=timestamps,
                                      ylab=frame_labels[i],
                                      ax=ax[i],
                                      **kwargs)
        elif style == "boxplot":
            time_series_boxplot(list_of_curves[i],
                                timestamps=timestamps,
                                ylab=frame_labels[i],
                                ax=ax[i],
                                **kwargs)
        else:
            raise Exception(f"Unrecognized style {style}")

        if actuals is not None:
            ax[i].scatter(timestamps, actuals,
                          s=cex_actuals * 100 / np.sqrt(len(actuals)))

        if vertical_cuts is not None:
            if vertical_cuts[i] == vertical_cuts[i]:
                ax[i].axvline(vertical_cuts[i])

    if main:
        fig.suptitle(main)
    return fig


def hosmer_lemeshow_plot(actual, predicted, ax=None, **kwargs):
    """
    Construct a Hosmer Lemeshow plot on the supplied axes.  A Hosmer Lemeshow
    plot partitions the predicted values into groups, and compares the midpoint
    of each group with the observed proportion in the group.

    Args:
      actual: The actual set of 0's and 1's being modeled.
      predicted: The predicted probabilities coresponding to 'actual'.
      ax: The pyplot axes on which to draw the plot, or None.  If None the
        figure will be shown on function exit.
      **kwargs: Passed to plt.subplots in the event that ax is None.

    Return:
      ax: The axes object containing the plot.
      group_means: The pd.Series containing the group means.  The series is
        indexed by pd.Interval objects indicating the interval over which the
        means are averaged.
    """
    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = None
    group_means = pd.DataFrame({"pred": predicted, "actual": actual}).groupby(
        pd.qcut(predicted, 10))["actual"].mean()
    bar_locations = group_means.index.categories.mid.values

    lower = np.array([x.left for x in group_means.index.values])
    upper = np.array([x.right for x in group_means.index.values])
    bar_widths = 1.0 * (upper - lower)

    plot_options, kwargs = _skim_plot_options(**kwargs)

    ax.bar(bar_locations, group_means, width=bar_widths, edgecolor="black",
           **kwargs)
    # ax.set_xlim(0, 1)
    # ax.set_ylim(0, 1)

    # labels = [str(lab) for lab in group_means.index.values]
    # ax.set_yticklabels(labels)
    ax.set_ylabel("Observed Proportions")

    # xticks = pretty_plot_ticks(np.min(predicted), np.max(predicted), 5)
    # ax.set_xticks(xticks)
    ax.set_xlabel("Predicted Proportions")

    _set_plot_options(ax, **plot_options)

    abline(a=0, b=1, ax=ax, color="black")

    if fig is not None:
        fig.show()

    return ax, group_means


def lines_gaussian_kde(kde, ax=None, **kwargs):
    """
    Add a kernel density estimate to the plot.
    """
    if ax is None:
        fig, ax = plt.subplots(1, 1)
    xlim = ax.get_xlim()
    x = np.linspace(xlim[0], xlim[1])
    y = kde.pdf(x)
    ax.plot(x, y, **kwargs)
    return ax


def lty(style):
    """
    Python linestyle characters from R 'lty' (linetype).

    Args:
      style:  An int or string.

    """
    style_names = ["solid", "dashed", "dotted", "dotdash",
                   "longdash", "twodash"]

    mappings = {
        "solid": "-",
        "dashed": "--",
        "dotted": ":",
        "dotdash": "-.",
        "longdash": (5, (8, 1, 8, 1)),
        "twodash": (5, (6, 1, 3, 1))
    }

    if isinstance(style, str):
        if style == "dashdot":
            style = "dotdash"
        style = unique_match(style, style_names)
        return mappings[style]

    elif isinstance(style, Number):
        style_number = style % len(style_names)
        return mappings[style_names[style_number]]



import BayesBoom.boom as boom
import numpy as np
from .state_models import StateModel
import BayesBoom.R as R


class LocalLinearTrendStateModel(StateModel):
    """
    The local linear trend model extends the local level model by including a
    slope parameter 'delta'.  The model is

             y[t] = mu[t] + error[t]
          mu[t+1] = mu[t] + delta[t] + innovation0[t]
       delta[t+1] = delta[t] + innovation1[t]

    The level component is "mu".  At each time point the mean changes by an
    expected amount delta (which is also time varying).  In this sense delta is
    the "local slope".

    The parameters of this model are the standard deviations of the innovation
    terms (which are modeled as independent zero-mean Gaussians).
    """

    def __init__(self,
                 y=None,
                 level_sigma_prior: R.SdPrior = None,
                 slope_sigma_prior: R.SdPrior = None,
                 initial_level_prior: R.NormalPrior = None,
                 initial_slope_prior: R.NormalPrior = None,
                 sdy: float = None,
                 initial_y: float = None):
        """
        Args:
          y:  The time series to be modeled.
          level_sigma_prior: Prior distribution on the standard deviation of
            the "level" innovations.
          slope_sigma_prior: Prior distribution on the standard deviation of
            the "slope" innovations.
          initial_level_prior: Prior distribution on the value of the level at
            time 0.
          initial_slope_prior: Prior distribution on the value of the slope at
            time 0.
          sdy:  The standard deviation of the time series being modeled.
          initial_y: The value of y at time 0.

        Details:
          There are three ways to build this object.
          (1) Pass all four prior objects.
          (2) Pass zero or more prior objects, sdy, and initial_y.
          (3) Pass y.

          Option 3 is the most convenient.  Option 1 offers the most control.
        """
        self._validate_priors(level_sigma_prior, slope_sigma_prior, y, sdy)
        self._validate_initial_distributions(
            initial_level_prior, initial_slope_prior, y, sdy, initial_y)
        self._state_contribution = None
        self._build_state_model()

    def __repr__(self):
        return (
            "A LocalLinearTrendStateModel with level_sd = "
            f"{self._state_model.sigma_level} and slope_sd = "
            f"{self._state_model.sigma_slope}."
        )

    @property
    def label(self):
        return "Trend (local linear)"

    @property
    def state_dimension(self):
        return 2

    @property
    def state_error_dimension(self):
        return 2

    @property
    def state_contribution(self):
        return self._state_contribution

    def allocate_space(self, niter, time_dimension):
        self.sigma_level = np.empty(niter)
        self.sigma_slope = np.empty(niter)
        self._state_contribution = np.empty((niter, time_dimension))

    def record_state(self, iteration, state_matrix):
        self.sigma_level[iteration] = self._state_model.sigma_level
        self.sigma_slope[iteration] = self._state_model.sigma_slope
        self._state_contribution[iteration, :] = state_matrix[
            self._state_index, :]

    def restore_state(self, iteration):
        self._state_model.set_sigma_level(self.sigma_level[iteration])
        self._state_model.set_sigma_slope(self.sigma_slope[iteration])

    def _validate_priors(self, level_sigma_prior, slope_sigma_prior, y, sdy):
        if level_sigma_prior is None:
            sdy = self._compute_sdy(sdy, y, "level_sigma_prior")
            level_sigma_prior = R.SdPrior(
                sigma_guess=.01 * sdy,
                upper_limit=sdy)
        if not isinstance(level_sigma_prior, R.SdPrior):
            raise Exception("Unexpected type for level_sigma_prior.")

        if slope_sigma_prior is None:
            sdy = self._compute_sdy(sdy, y, "slope_sigma_prior")
            slope_sigma_prior = R.SdPrior(
                sigma_guess=.01 * sdy,
                upper_limit=sdy)
        if not isinstance(slope_sigma_prior, R.SdPrior):
            raise Exception("Unexpected type for slope_sigma_prior.")

        self._level_sigma_prior = level_sigma_prior
        self._slope_sigma_prior = slope_sigma_prior

    def _validate_initial_distributions(
            self, initial_level_prior, initial_slope_prior,
            y, sdy, initial_y):
        if initial_level_prior is None:
            sdy = self._compute_sdy(sdy, y, "initial_level_prior")
            if initial_y is None:
                if y is None:
                    raise Exception(
                        "One of initial_y, y, or initial_level_prior must be "
                        "specified.")
                else:
                    initial_y = y[0]
            initial_level_prior = R.NormalPrior(initial_y, sdy)
        if not isinstance(initial_level_prior, R.NormalPrior):
            raise Exception("Unexpected type for initial_level_prior.")
        self._initial_level_prior = initial_level_prior

        if initial_slope_prior is None:
            sdy = self._compute_sdy(sdy, y, "initial_slope_prior")
            initial_slope_prior = R.NormalPrior(0, sdy)
        if not isinstance(initial_slope_prior, R.NormalPrior):
            raise Exception("Unexpected type for initial_slope_prior.")
        self._initial_slope_prior = initial_slope_prior

    def _build_state_model(self):
        self._state_model = boom.LocalLinearTrendStateModel()
        self._set_posterior_sampler()
        self._set_initial_distribution()

    def _set_posterior_sampler(self):
        """
        A utility called by the constructor.  See the __init__ method for
        argument documentation.
        """

        self._state_model.set_posterior_sampler(
            self._level_sigma_prior.create_chisq_model(),
            self._level_sigma_prior.upper_limit,
            self._slope_sigma_prior.create_chisq_model(),
            self._slope_sigma_prior.upper_limit,
            boom.GlobalRng.rng)

    def _set_initial_distribution(self):
        """
        A utility called by the constructor.  See the __init__ method for
        argument documentation.
        """
        initial_state_mean = np.array([
            self._initial_level_prior.mean,
            self._initial_slope_prior.mean,
        ])
        self._state_model.set_initial_state_mean(
            boom.Vector(initial_state_mean))
        initial_sd = np.array([
            self._initial_level_prior.sd,
            self._initial_slope_prior.sd,
        ])
        initial_variance = np.diag(initial_sd * initial_sd)
        self._state_model.set_initial_state_variance(
            boom.SpdMatrix(initial_variance))

    @staticmethod
    def _compute_sdy(sdy, y, which_prior):
        """
        Return the standard deviation of y, computing it if and only if needed.
        """
        if sdy is None:
            if y is None:
                raise Exception(
                    f"One of y, sdy, or {which_prior} must be specified.")
            else:
                sdy = np.nanstd(y, ddof=1)
        return sdy

    def __getstate__(self):
        payload = self.__dict__.copy()
        del payload["_state_model"]
        return payload

    def __setstate__(self, payload):
        self.__dict__ = payload
        self._build_state_model()



import os
import pathlib


def find_project_root(dir=None):
    """
    Return the first enclosing directory containing a file named 'setup.py'
    """
    if dir is None:
        dir = os.path.curdir

    if os.path.exists(os.path.join(str(dir), "setup.py")):
        return dir
    else:
        current_path = pathlib.path(dir)
        parent = current_path.parent.absolute()
        if parent == current_path:
            return current_path
        else:
            return find_project_root(parent)



import unittest
import BayesBoom.R as R
import numpy as np
import pandas as pd
import json


class TestEffectEncoder(unittest.TestCase):

    def setUp(self):
        np.random.seed(8675309)

    def test_init(self):
        encoder = R.EffectEncoder("Color", ["Red", "Blue", "Green"])
        self.assertEqual(encoder._levels, ["Red", "Blue"])
        self.assertEqual(encoder._baseline, "Green")
        self.assertEqual(2, encoder.dim)

        self.assertEqual(["Color.Red", "Color.Blue"],
                         encoder.encoded_variable_names)

        encoder_rep = str(encoder)
        self.assertEqual("EffectEncoder for Color", encoder_rep)

    def test_encoding(self):
        encoder = R.EffectEncoder("Color", ["Red", "Blue", "Green"])
        x = ["Red", "Blue", "Green"]
        enc = encoder.encode(x)
        expected = np.array([[1.0, 0.0],
                             [0.0, 1.0],
                             [-1.0, -1.0]])
        self.assertTrue(np.allclose(enc, expected))

    def test_new_level(self):
        encoder = R.EffectEncoder("Color", ["Red", "Blue", "Green"])
        enc = encoder.encode(["Red", "Blue", "Orange", np.NaN, 17, "Green"])
        expected = np.array([[1.0, 0.0],
                             [0.0, 1.0],
                             [0.0, 0.0],
                             [0.0, 0.0],
                             [0.0, 0.0],
                             [-1.0, -1.0]])
        self.assertTrue(np.allclose(enc, expected))

    def test_encode_dataset(self):
        data = pd.DataFrame(np.random.randn(3, 2),
                            columns=["X1", "X2"])
        data["Color"] = ["Red", "Blue", "Green"]
        encoder = R.EffectEncoder("Color", ["Red", "Blue", "Green"])
        enc = encoder.encode_dataset(data)
        expected = np.array([
            [1.0, 0.0],
            [0.0, 1.0],
            [-1.0, -1.0]])
        self.assertTrue(np.allclose(enc, expected))

    def test_json(self):
        encoder = R.EffectEncoder("Color", ["Red", "Blue", "Green"])
        data = ["Red", "Blue", "Orange", np.NaN, 17, "Green"]
        enc = encoder.encode(data)
#         json_encoder = R.MainEffectEncoderJsonEncoder()
        json_string = json.dumps(encoder,
                                 cls=R.MainEffectEncoderJsonEncoder)
        encoder2 = json.loads(json_string, cls=R.MainEffectEncoderJsonDecoder)
        enc2 = encoder2.encode(data)
        self.assertTrue(np.allclose(enc, enc2))


class TestOneHotEncoder(unittest.TestCase):
    def setUp(self):
        np.random.seed(8675309)

    def test_encoding(self):
        levels = ["Red", "Blue", "Green"]
        encoder = R.OneHotEncoder("blah", levels, baseline_level="Green")
        self.assertEqual(2, encoder.dim)
        x = ["Red", "Red", "Green", "Blue"]
        X = encoder.encode(x)
        self.assertEqual((4, 2), X.shape)
        expected = np.array([
            [1.0, 0.0],
            [1.0, 0.0],
            [0.0, 0.0],
            [0.0, 1.0]
        ])
        self.assertTrue(np.allclose(X, expected))

    def test_json(self):
        levels = ["Red", "Blue", "Green"]
        encoder = R.OneHotEncoder("blah", levels, baseline_level="Green")
        data = ["Red", "Red", "Green", "Blue"]
        output = encoder.encode(data)

        json_string = json.dumps(encoder, cls=R.MainEffectEncoderJsonEncoder)
        encoder2 = json.loads(json_string, cls=R.MainEffectEncoderJsonDecoder)
        self.assertTrue(np.allclose(output, encoder2.encode(data)))


class TestIdentityEncoder(unittest.TestCase):
    def setUp(self):
        np.random.seed(8675309)

    def test_everything(self):
        encoder = R.IdentityEncoder("Blah")
        self.assertEqual(encoder.variable_name, "Blah")
        x = np.random.randn(4)
        output = encoder.encode(x)
        self.assertEqual(output.shape, (4, 1))
        self.assertTrue(np.allclose(output.ravel(), x))

        json_string = json.dumps(encoder, cls=R.MainEffectEncoderJsonEncoder)
        encoder2 = json.loads(json_string, cls=R.MainEffectEncoderJsonDecoder)
        self.assertTrue(np.allclose(output, encoder2.encode(x)))


class TestSuccessEncoder(unittest.TestCase):
    def setUp(self):
        np.random.seed(8675309)

    def test_encoding(self):
        encoder = R.SuccessEncoder("Blah", ["A", "C"])
        data = ["A", "B", "C", "D"]
        enc = encoder.encode(data)
        self.assertEqual(enc.shape, (4, 1))
        self.assertTrue(np.allclose(
            enc.ravel(), np.array([1, 0, 1, 0])))

    def test_json(self):
        encoder = R.SuccessEncoder("Blah", ["A", "C"])
        data = ["A", "B", "C", "D"]
        enc = encoder.encode(data)

        json_string = json.dumps(encoder, cls=R.MainEffectEncoderJsonEncoder)
        encoder2 = json.loads(json_string, cls=R.MainEffectEncoderJsonDecoder)
        self.assertTrue(np.allclose(enc, encoder2.encode(data)))


class TestDatasetEncoder(unittest.TestCase):
    def setUp(self):
        np.random.seed(8675309)

    def test_init(self):
        enc1 = R.EffectEncoder("Color", ["Red", "Blue"])
        enc2 = R.IdentityEncoder("Height")
        enc3 = R.InteractionEncoder(enc1, enc2)
        d = R.DatasetEncoder([enc1, enc2, enc3])
        self.assertIsInstance(d, R.DatasetEncoder)

    def test_encoding(self):
        enc1 = R.EffectEncoder("Color", ["Red", "Blue"])
        enc2 = R.IdentityEncoder("Height")
        enc3 = R.InteractionEncoder(enc1, enc2)
        encoder = R.DatasetEncoder([enc1, enc2, enc3])

        sample_size = 1000
        data = pd.DataFrame(
            {
                "Height": np.random.randn(sample_size),
                "Color": np.random.choice(["Red", "Blue"], sample_size)
            }
        )
        enc = encoder.encode_dataset(data)
        self.assertEqual(sample_size, enc.shape[0])
        self.assertEqual(4, enc.shape[1])
        self.assertTrue(np.allclose(enc[:, 2], data.iloc[:, 0]))

        json_string = json.dumps(encoder, cls=R.DatasetEncoderJsonEncoder)
        encoder2 = json.loads(json_string, cls=R.DatasetEncoderJsonDecoder)
        self.assertTrue(np.allclose(enc, encoder2.encode_dataset(data)))


_debug_mode = False

if _debug_mode:
    import pdb  # noqa

    # Turn warnings into errors.
    # warnings.simplefilter("error")

    # Run the test you are trying to debug here.  Instantiate the test class,
    # then call the problematic test.  Call pdb.pm() in the event of an
    # exception.
    print("Hello, world!")

    rig = TestDatasetEncoder()
    if hasattr(rig, "setUpClass"):
        rig.setUpClass()
    if hasattr(rig, "setUp"):
        rig.setUp()

    rig.test_encoding()

    print("Goodbye, cruel world!")

else:
    if __name__ == "__main__":
        unittest.main(verbosity=2)



import BayesBoom.boom as boom
import numpy as np
import BayesBoom.R as R
from .state_models import StateModel


class LocalLevelStateModel(StateModel):
    """
    The local level model assumes the data move as a noisy random walk.
              y[t] = mu[t] + error[t], mu[t+1] = mu[t] + innovation[t]
     innovation[t] ~ N(0, sigsq)

    The model parameter is the variance of the innovation terms.
    """

    def __init__(self, y, sigma_prior=None, initial_state_prior=None,
                 sdy=None, initial_y=None):
        """
        Args:
          y: The data to be modeled.  If sdy and initial_y are supplied
            this is not used.
          sigma_prior: An object of class boom.GammaModelBase serving as the
            prior on the precision (reciprocal variance) of the innovation
            terms.  If None then 'sdy' will be used to choose a defalt.
          initial_state_prior: An object of class boom.GaussianModel serving as
            the prior distribution on the value of the state at time 0 (the
            time of the first observation).  If None then initial_y and sdy
            will be used to choose a defalt.
          sdy: The standard deviation of y.  If None then this will be computed
            from y.  This argument is primarily intended to handle unusual
            cases where 'y' is unavailable.
          initial_y: The first element of y.  If None then this will be
            computed from y.  This argument is primarily intended to handle
            unusual cases where 'y' is unavailable.

        Returns:
          A StateModel object representing a local level model.
        """
        self._validate_sigma_prior(sigma_prior, y, sdy)
        self._validate_initial_distributions(initial_state_prior, y, sdy,
                                             initial_y)
        self._build_state_model()
        self._state_contribution = None

    def __repr__(self):
        return f"Local level with sigma = {self._state_model.sigma}"

    @property
    def label(self):
        return "Trend (local level)"

    @property
    def state_dimension(self):
        return 1

    def allocate_space(self, niter, time_dimension):
        self.sigma_draws = np.zeros(niter)
        self._state_contribution = np.zeros((niter, time_dimension))

    def record_state(self, iteration, state_matrix):
        self.sigma_draws[iteration] = self._state_model.sigma
        if self._state_index < 0:
            raise Exception("Each state model must be told where its state"
                            "component begins in the global state vector.  "
                            "Try calling set_state_index.")
        self._state_contribution[iteration, :] = state_matrix[
            self._state_index, :]

    def restore_state(self, iteration):
        self._state_model.set_sigma(self.sigma_draws[iteration])

    @property
    def state_contribution(self):
        return self._state_contribution

    def _validate_sigma_prior(self, sigma_prior, y, sdy):
        if sigma_prior is None:
            if sdy is None:
                sdy = np.nanstd(y, ddof=1)
            sigma_prior = R.SdPrior(sigma_guess=.01 * sdy,
                                    sample_size=.01,
                                    upper_limit=sdy)
            if not isinstance(sigma_prior, R.SdPrior):
                raise Exception("sigma_prior should be an R.SdPrior.")
        self._sigma_prior = sigma_prior

    def _validate_initial_distributions(self, initial_state_prior, y, sdy,
                                        initial_y):
        if initial_state_prior is None:
            if initial_y is None:
                initial_y = y[0]
            if sdy is None:
                sdy = np.nanstd(y, ddof=1)
            initial_state_prior = R.NormalPrior(float(initial_y), float(sdy))
        if not isinstance(initial_state_prior, R.NormalPrior):
            raise Exception(
                "initial_state_prior should be an R.NormalPrior.")
        self._initial_state_prior = initial_state_prior

    def _build_state_model(self):
        self._state_model = boom.LocalLevelStateModel()
        self._state_model.set_initial_state_mean(self._initial_state_prior.mean)
        self._state_model.set_initial_state_variance(
            self._initial_state_prior.variance)
        self._state_model.set_sigma(self._sigma_prior.initial_value)
        state_model_sampler = self._state_model.set_posterior_sampler(
            self._sigma_prior.boom())
        state_model_sampler.set_sigma_upper_limit(
            self._sigma_prior.upper_limit)

    def __getstate__(self):
        payload = self.__dict__.copy()
        del payload["_state_model"]
        return payload

    def __setstate__(self, payload):
        self.__dict__ = payload
        self._build_state_model()



import scipy.stats as sp
import numpy as np

# expose R's equivalent of pnorm, qnorm, etc.


def as_numeric(x):
    return np.array(x)

def dnorm(x, mean=0, sd=1, log=False):
    """Normal density function.

    :param x: number, list or numpy array
        The argument of the density function.

    :param x: number, list or numpy array
        The argument of the density function.

    """
    x = as_numeric(x)
    if log:
        return sp.norm.logpdf(x, mean, sd)
    else:
        return sp.norm.pdf(x, mean, sd)


def pnorm(x, mean=0, sd=1, lower=True, log=False):
    """
    Normal cumulative distribuion function.

    """
    if log:
        if lower:
            return sp.norm.logcdf(x, loc=mean, scale=sd)
        else:
            return sp.norm.logsf(x, loc=mean, scale=sd)
    else:
        if lower:
            return sp.norm.cdf(x, loc=mean, scale=sd)
        else:
            return sp.norm.sf(x, loc=mean, scale=sd)


def qnorm(x, mean=0, sd=1, lower=True, log=False):
    """
    Quantiles of the normal distribuion.

    """
    if log:
        x = np.exp(x)

    if lower:
        return sp.norm.ppf(x, mean, sd)
    else:
        return sp.norm.ppf(1 - x, mean, sd)


def rnorm(n, mean=0, sd=1):
    """
    Random deviates from the normal distribution.
    """

    return np.random.randn(n) * sd + mean


def rbeta(n, a=1, b=1):
    from scipy.stats import beta
    return beta.rvs(a, b, size=n)


def rpois(n, lam):
    from scipy.stats import poisson
    return poisson.rvs(lam, size=n)


def rbinom(n, size, prob):
    from scipy.stats import binom
    return binom.rvs(size, prob, size=n)


def dgamma(y, shape, scale, log=False):
    from scipy.stats import gamma
    if log:
        return gamma.logpdf(y, shape, scale=1.0 / scale)
    else:
        return gamma.pdf(y, shape, scale=1.0 / scale)


def pgamma(y, shape, scale, lower_tail=True, log=False):
    from scipy.stats import gamma
    if lower_tail:
        if log:
            return gamma.logcdf(y, shape, scale=1.0 / scale)
        else:
            return gamma.cdf(y, shape, scale=1.0 / scale)
    else:
        if log:
            return gamma.logsf(y, shape, scale=1.0 / scale)
        else:
            return gamma.sf(y, shape, scale=1.0 / scale)


def qgamma(probs, shape, scale, lower_tail=True, log=False):
    from scipy.stats import gamma
    if log:
        probs = np.exp(probs)
    if lower_tail:
        return gamma.ppf(probs, shape, scale=1.0 / scale)
    else:
        return gamma.isf(probs, shape, scale=1.0 / scale)


def rgamma(n, shape, scale):
    """
    Random deviates from the gamma distribution with mean shape / scale.
    """
    from scipy.stats import gamma
    return gamma.rvs(shape, size=n, scale=1.0 / scale)


def rmarkov(n: int, P: np.ndarray, pi0=None):
    """
    Simulation a Markov chain of length n from transition probability matrix P,
    with initial distribution pi0.  The state space is the set of integers in
    {0, ..., S-1} where S is the number of rows in P.

    Args:
      n:  The length of the Markov chain to simulate.
      P: A transition probability matrix.  A square matrix of non-negative
         elements, where each row sums to 1.
      pi0: The distribution of the state at time 0.  A vector of non-negative
        numbers summing to 1.

    Returns:
      A numpy array of integer dtype containing the simulated Markov chain..
    """
    P = np.array(P)
    if len(P.shape) != 2 or P.shape[0] != P.shape[1]:
        raise Exception("P must be a square matrix.")
    S = P.shape[0]
    if pi0 is None:
        pi0 = np.ones(S) / S
    else:
        pi0 = np.array(pi0).flatten()

    if len(pi0) != S:
        raise Exception("Initial distribution must have the same "
                        "dimension as the transition probability matrix.")
    ans = np.full(n, -1)
    ans[0] = np.random.choice(range(S), p=pi0)
    for i in range(1, n):
        ans[i] = np.random.choice(range(S), p=P[ans[i-1], :])
    return(ans)


def dmvn(y, mu, Sigma, inv=False, logscale=False):
    """
    Multivariate normal density.

    Args:
      y: Either a vector or a 2-D array.  If a vector is passed then the output
        is a scalar density value.  If a matrix is passed, the return value is
        a vector with each element giving the density for the corresponding row
        of y.
      mu: If y is a vector, mu must be a vector of the same dimension.  If y is
       a matrix, then mu is either a vector with lenght matching the number of
       columns in y, or a matrix with the same shape as y.
      Sigma: If y is a vector Sigma is a symmetric positive definite matrix
        matching the length of y.  If y is a matrix then either Sigma is a
        single matrix matching the length of y, or a 3-way array with first
        dimension matching the first dimension of y, and Sigma[i, :, :] giving
        the variance of y[i, :].
      inv: If True then Sigma represents the precision of y (the inverse
        variance).  If False then Sigma represents the variance.
      logscale: If True then the log of the density is returned.  If False then
        the regular density value is returned.
    """

    y = np.array(y)
    if len(y.shape) > 2:
        raise Exception("dmvn requires either a vector or matrix input for y.")
    elif len(y.shape) == 1:
        y = y.reshape((1, -1))
    nobs = y.shape[0]
    dim = y.shape[1]

    mu = np.array(mu)
    if len(mu.shape) > 2:
        raise Exception("dmvn requires either a vector or matrix input for mu.")
    if len(mu.shape) == 1:
        mu = mu.reshape((1, -1))
    if mu.shape[0] == 1 and nobs > 1:
        mu = np.array([mu.ravel()] * nobs)
    if mu.shape[0] != nobs:
        raise Exception("The shapes of mu and y must match.")

    if len(Sigma.shape) > 4:
        raise Exception(
            "dmvn requires either a matrix or 3-way array for Sigma")
    if len(Sigma.shape) < 2:
        raise Exception("Sigma can't be a vector in dmvn.")
    if len(Sigma.shape) == 2:
        if not np.allclose(Sigma, Sigma.T):
            raise Exception(
                "Sigma must be a symmetric positive definite matrix.")
        if not inv:
            Sigma = np.linalg.inv(Sigma)
            inv = True
        ldsi = np.array([np.linalg.slogdet(Sigma)[1]] * nobs)
        Sigma = np.array([Sigma] * nobs)
    if len(Sigma.shape) != 3:
        raise Exception("Something with wrong with Sigma")
    if Sigma.shape[0] == 1 and nobs > 1:
        if not inv:
            Sigma = np.linalg.inverse(Sigma[0, :, :])
            inv = True
        else:
            Sigma = Sigma[0, :, :]
        ldsi = np.array([np.linalg.slogdet(Sigma)] * nobs)[1]
        Sigma = np.array([Sigma] * nobs)
    if Sigma.shape[0] != nobs or Sigma.shape[1] != dim or Sigma.shape[2] != dim:
        raise Exception("The shapes of y and Sigma must match.")
    if not inv:
        Sigma = np.linalg.inv(Sigma)
        ldsi = np.linalg.slogdet(Sigma)[1]

    log2pi = 1.83787706641
    residual = y - mu

    qform = np.einsum('ij,ijk,ik -> i', residual, Sigma, residual)

    ans = 0.5 * (-dim * log2pi + ldsi - qform)
    if nobs == 1:
        ans = float(ans)
    if logscale:
        return ans
    return np.exp(ans)


def rmvn(n, mu, Sigma, drop=True):
    """
    Draws from the multivariate normal distribution with mean mu and variance
    matrix Sigma.

    Args:

      n: The number of desired draws.

      mu: A numpy vector or matrix giving the mean of the draws.  If a vector
        then mu is the mean for all .  If a matrix then mu[i, :] is the mean
        vector for draw i.

      Sigma: Either a 2D or a 3D numpy array.  If a 2D array is passed then
        Sigma is the common variance matrix used for all draws.  If a 3D array
        is passed, then Sigma[i, :, :] is the variance matrix for draw i.
    """
    mu = np.array(mu)
    if len(mu.shape) > 2:
        raise Exception("dmvn requires either a vector or matrix input for mu.")
    if len(mu.shape) == 1:
        mu = mu.reshape((1, -1))
    if mu.shape[0] == 1 and n > 1:
        mu = np.array([mu.ravel()] * n)
    if mu.shape[0] != n:
        raise Exception(f"Requested {n} draws but passed a 'mu' argument with"
                        f" {mu.shape[0]} rows.")

    dim = mu.shape[1]
    Z = np.random.randn(n, dim)

    if len(Sigma.shape) == 2:
        L = np.linalg.cholesky(Sigma)
        draws = (L @ Z.T).T + mu
    elif len(Sigma.shape) != 3:
        raise Exception(
            "Either a matrix or a 3-way array is required for Sigma")
    else:
        draws = np.array([
            np.linalg.cholesky(Sigma[i, :, :]) @ Z[i, :] + mu[i, :]
            for i in range(n)])

    if n == 1 and drop:
        return draws[0, :]
    else:
        return draws


def rmulti(probs, n=1):
    """
    Simulate one or more draws from the given discrete probability
    distribution.
    """
    return np.random.choice(range(len(probs)), size=n, replace=True, p=probs)



import pandas as pd
from pandas.api.types import is_numeric_dtype


def to_data_table(data: pd.DataFrame):
    """
    Convert a pandas DataFrame to a BOOM DataTable.
    """
    import BayesBoom.boom as boom
    table = boom.DataTable()
    for vname in data.columns:
        if is_numeric_dtype(data[vname]):
            table.add_numeric(boom.Vector(data[vname].values.astype("float")),
                              str(vname))
        else:
            variable = data[vname].astype("category")
            categories = variable.cat.categories.tolist()
            table.add_categorical(
                variable.cat.codes.values,
                [str(x) for x in categories],
                vname,
            )
    return table


def to_pandas(data):
    """
    Convert a Boom.boom.DataTable to a pandas DataFrame.
    """
    ncol = data.ncol
    variables = {}
    vnames = data.variable_names
    for i in range(ncol):
        dtype_str = data.variable_type(i)
        if dtype_str == "numeric":
            variables[vnames[i]] = pd.Series(data.getvar(i), dtype="float")
        elif dtype_str == "categorical":
            codes = data.get_nominal_values(i)
            categories = data.get_nominal_levels(i)
            variables[vnames[i]] = pd.Series(pd.Categorical.from_codes(
                codes, categories))
    return pd.DataFrame(variables)



import numpy as np
import pandas as pd
from numbers import Number


def _cbind_numpy(*args):
    """
    Collect all the arguments together in a single Matrix.  Scalar arguments
    are repeated to the implied number of rows.
    """
    args = [*args]
    shapes = [x.shape for x in args if not isinstance(x, Number)]
    if len(shapes) == 0:
        raise Exception("At least one array must be passed to cbind.")
    nrows = np.array([x[0] for x in shapes])
    if not np.all(nrows == nrows[0]):
        raise Exception("Incompatible shapes in cbind.")
    nrows = nrows[0]
    ncols = [x[1] for x in shapes if len(x) > 1]
    args = [np.full(nrows, x) if isinstance(x, Number) else x
            for x in args]
    if len(ncols) == 0:
        return np.array(args).T
    else:
        return np.hstack([x.reshape((nrows, -1)) for x in args])


def _cbind_pandas(*args):
    """
    Collect all the arguments together in a single data frame.  Scalar
    arguments are repeated to the implied number of rows.
    """
    args = [*args]
    shapes = [x.shape for x in args if not isinstance(x, Number)]
    if len(shapes) == 0:
        raise Exception("At least one array must be passed to cbind.")
    nrows = np.array([x[0] for x in shapes])
    if not np.all(nrows == nrows[0]):
        raise Exception("Incompatible shapes in cbind.")
    nrows = nrows[0]

    names = []
    pos = 0
    for i, arg in enumerate(args):
        if isinstance(arg, Number):
            names.append(f"X{pos + 1}")
            pos += 1
        elif isinstance(arg, np.ndarray):
            if len(arg.shape) == 1:
                names.append(f"X{pos + 1}")
                pos += 1
            elif len(arg.shape) == 2:
                names += [f"X{pos + i + 1}" for i in range(arg.shape[1])]
                pos += arg.shape[1]
            else:
                raise Exception(
                    "Can't call cbind on arrays with dimension more than 2")
        elif isinstance(arg, pd.Series):
            names.append(f"X{pos + 1}")
            pos += 1
        elif isinstance(arg, pd.DataFrame):
            names += list(arg.columns)

    args = [np.full(nrows, x) if isinstance(x, Number) else x
            for x in args]
    args = [pd.DataFrame(x) for x in args]

    ans = pd.concat(args, axis=1, ignore_index=True)
    ans.columns = names
    return ans


def cbind(*args):
    args = [*args]
    has_pandas = np.any([
        isinstance(x, pd.Series)
        or isinstance(x, pd.DataFrame)
        for x in args])
    if has_pandas:
        return _cbind_pandas(*args)
    else:
        return _cbind_numpy(*args)



import unittest
import BayesBoom.R as R
from BayesBoom.R import dmvn, rmvn
import numpy as np
import scipy.stats as ss


class TestDmvn(unittest.TestCase):

    def setUp(self):
        np.random.seed(8675309)

    def test_eval(self):
        Z = np.random.randn(10, 3)
        Sigma = Z.T @ Z
        Sigma_chol = np.linalg.cholesky(Sigma)

        mu = np.random.randn(3)

        y = np.random.randn(5, 3)
        for i in range(5):
            y[i, :] = Sigma_chol @ y[i, :] + mu

        prob = dmvn(y, mu, Sigma)
        obj = ss.multivariate_normal(mean=mu, cov=Sigma)
        self.assertTrue(np.allclose(prob, obj.pdf(y)))

        # Test logscale
        logprob = dmvn(y, mu, Sigma, logscale=True)
        self.assertTrue(np.allclose(logprob, obj.logpdf(y)))

        # Test with parameter arrays.
        Z = np.random.randn(10, 3)
        Sigma2 = Z.T @ Z

        Z = np.random.randn(10, 3)
        Sigma3 = Z.T @ Z
        Sigma_array = np.array([Sigma, Sigma2, Sigma3])

        prob = dmvn(y[:3, :], mu, Sigma_array)
        self.assertAlmostEqual(prob[0], dmvn(y[0, :], mu, Sigma_array[0, :, :]))
        self.assertAlmostEqual(prob[1], dmvn(y[1, :], mu, Sigma_array[1, :, :]))
        self.assertAlmostEqual(prob[2], dmvn(y[2, :], mu, Sigma_array[2, :, :]))

        mu_array = np.random.randn(4, 3)
        prob = dmvn(y[:4, :], mu_array, Sigma)
        self.assertAlmostEqual(prob[0], dmvn(y[0, :], mu_array[0, :], Sigma))
        self.assertAlmostEqual(prob[1], dmvn(y[1, :], mu_array[1, :], Sigma))
        self.assertAlmostEqual(prob[2], dmvn(y[2, :], mu_array[2, :], Sigma))
        self.assertAlmostEqual(prob[3], dmvn(y[3, :], mu_array[3, :], Sigma))


class TestRmvn(unittest.TestCase):
    def setUp(self):
        np.random.seed(8675309)

    def test_draws(self):
        mu = np.array([1, 3, 7])
        draws = rmvn(10, mu, np.diag(np.ones(3)))
        self.assertEqual(draws.shape, (10, 3))

        draws = rmvn(10000, mu, np.diag(np.ones(3)))
        epsilon = 5 * 1.0 / np.sqrt(draws.shape[0])
        self.assertTrue(np.all(np.abs(draws.mean(axis=0) - mu) < epsilon))

        Sigma = np.array([
            [1.0, 0.4, -.3],
            [0.4, 2.0, 0.9],
            [-.3, 0.9, 8.0]])

        draws = rmvn(100000, mu, Sigma)
        epsilon = 5 * np.sqrt(np.diag(Sigma) / draws.shape[0])
        self.assertTrue(np.all(np.abs(draws.mean(axis=0) - mu) < epsilon))

        V = R.var(draws)
        self.assertTrue(np.all(np.abs(V - Sigma) < .05))

    def test_multiple_sigma_values(self):

        Sigma1 = np.array([
            [1.0, 0.4, -.3],
            [0.4, 2.0, 0.9],
            [-.3, 0.9, 8.0]])
        Sigma2 = 9 * Sigma1

        mu1 = np.array([1, 3, 7])
        mu2 = np.array([70, -30, 1])

        n1 = 1000
        n2 = 2000
        mu = np.array([mu1] * n1 + [mu2] * n2)
        Sigma = np.array([Sigma1] * n1 + [Sigma2] * n2)

        draws = R.rmvn(n1 + n2, mu, Sigma)
        self.assertEqual(draws.shape, (3000, 3))
        epsilon1 = 5 * np.sqrt(np.diag(Sigma1) / n1)
        self.assertTrue(np.all((draws[:n1, :].mean(axis=0) - mu1) < epsilon1))

        V1 = R.var(draws[:n1, :])
        relative_error = np.max(np.abs((V1 - Sigma1) / Sigma1))
        self.assertLess(relative_error, .2)

        epsilon2 = 5 * np.sqrt(np.diag(Sigma2) / n2)
        self.assertTrue(np.all((draws[n1:, :].mean(axis=0) - mu2) < epsilon2))
        V2 = R.var(draws[n1:, :])
        relative_error = np.max(np.abs((V2 - Sigma2) / Sigma2))
        self.assertLess(relative_error, .2)


_debug_mode = False

if _debug_mode:
    import pdb  # noqa

    # Turn warnings into errors.
    # warnings.simplefilter("error")

    # Run the test you are trying to debug here.  Instantiate the test class,
    # then call the problematic test.  Call pdb.pm() in the event of an
    # exception.
    print("Hello, world!")

    rig = TestRmvn()
    if hasattr(rig, "setUpClass"):
        rig.setUpClass()
    if hasattr(rig, "setUp"):
        rig.setUp()

    rig.test_multiple_sigma_values()

    print("Goodbye, cruel world!")

else:
    if __name__ == "__main__":
        unittest.main(verbosity=2)



import unittest
import BayesBoom.mixtures as mix
import numpy as np
import BayesBoom.R as R


class TestLabelSwitching(unittest.TestCase):
    def setUp(self):
        np.random.seed(8675309)

    def test_permutations(self):
        mu1 = np.zeros(3)
        mu2 = np.ones(3) * 4
        # mu2 = np.array([1, 10, 100])

        Sigma1 = np.diag(np.ones(3))
        Sigma2 = np.diag(np.ones(3))

        y1 = np.random.randn(100, 3) + mu1
        y2 = np.random.randn(150, 3) + mu2
        y = np.concatenate((y1, y2), axis=0)

        model = mix.DirichletProcessMvn(y)
        model.mcmc(100, ping=-1, permute_state_labels=False)

        true_cluster_probs1 = R.dmvn(y, mu1, Sigma1) * 100 / 250
        true_cluster_probs2 = R.dmvn(y, mu2, Sigma2) * 150 / 250
        total = true_cluster_probs1 + true_cluster_probs2

        true_cluster_probs = R.cbind(true_cluster_probs1 / total,
                                     true_cluster_probs2 / total)

        niter = 10
        nobs = y.shape[0]
        draws = np.empty((niter, nobs))
        perm = np.empty(niter)

        for i in range(niter):
            u = np.random.rand(nobs)
            draws[i, :] = u < true_cluster_probs[:, 1]
            v = np.random.rand(1)
            if v < .5:
                perm[i] = 1
                draws[i, :] = 1 - draws[i, :]
            else:
                perm[i] = 0

        fitted_perm = mix.identify_permutation_from_labels(draws)
        self.assertTrue(np.allclose(perm, fitted_perm[:, 1]))

    def test_choose_permutation(self):
        n1 = 100
        n2 = 200
        mu1 = np.zeros(3)
        mu2 = np.ones(3) * 4

        Sigma1 = np.diag(np.ones(3))
        Sigma2 = np.diag(np.ones(3))

        y1 = R.rmvn(n1, mu1, Sigma1)
        y2 = R.rmvn(n2, mu2, Sigma2)
        y = np.concatenate((y1, y2), axis=0)

        model = mix.DirichletProcessMvn(y)
        model.mcmc(1000, ping=-1, permute_state_labels=True)

        nscore1 = 3
        nscore2 = 5
        data_to_cluster = np.concatenate(
            (R.rmvn(nscore1, mu1, Sigma1),
             R.rmvn(nscore2, mu2, Sigma2)), axis=0)

        clusters = model.cluster(data_to_cluster, nclusters=2, prob=False)
        cluster1 = clusters[0]
        self.assertTrue(np.all(clusters[:nscore1] == cluster1))
        cluster2 = clusters[-1]
        self.assertTrue(np.all(clusters[nscore1:] == cluster2))

        cluster_probs = model.cluster(data_to_cluster, nclusters=2, prob=True)
        self.assertTrue(np.all(
            cluster_probs[:nscore1, cluster1] > cluster_probs[
                :nscore1, cluster2]))
        self.assertTrue(np.all(
            cluster_probs[nscore1:, cluster2] > cluster_probs[
                nscore1:, cluster1]))


_debug_mode = False

if _debug_mode:
    import pdb  # noqa

    # Run the test you are trying to debug here.  Instantiate the test class,
    # then call the problematic test.  Call pdb.pm() in the event of an
    # exception.
    print("Hello, world!")

    rig = TestLabelSwitching()

    if hasattr(rig, "setUpClass"):
        rig.setUpClass()
    if hasattr(rig, "setUp"):
        rig.setUp()

    rig.test_choose_permutation()

    print("Goodbye, cruel world!")

else:
    if __name__ == "__main__":
        unittest.main(verbosity=2)
